"""Heuristics for minimizing generator startup costs.

Functions:
minimize_start_costs
update_storage
avoid_generation
find_case
move_generation
alt_generation
spare_storage_check
leave_gen_on
seg_length
slack_cap
stor_state
turndown_capacity
turndown_now
stor_add
test_more_cases
mp_test_more_cases
"""

from eagers.basic.gen_limit import gen_limit, chp_heat
from eagers.update.update_cost import update_cost
from eagers.basic.marginal_cost import gen_marginal_cost, sort_mc, net_marginal_cost
from eagers.update.disable_gen import disable_gen
from eagers.solver.ecos_qp import ecos_qp
from eagers.solver.sort_solution import sort_gen_disp_step
from eagers.solver.dispatch_step import building_forecast_now
from eagers.basic.all_demands import all_demands
from eagers.basic.constrain_min_max import constrain_min_max
from eagers.update.update_matrices_step import update_matrices_step


def minimize_start_costs(qp, gen, market, subnet, options, forecast, first_profile, scale_cost, 
                        gen_output_w_starts, stor_power, binary_comb, disp_comb, build_temp, 
                        fluid_temp, cost_comb, verified, observer, dt, v_h):
    '''
    this function finds the unit commitment which minimizes startup costs, it relies on heuristic rules
    #the heuristic rules are present in the sub functions. It is called within dispatch_loop
    #this function finds the n shortest segments of time the generator is on/off, and compares the
    #startup cost to the alternative configuration that avoids the start or re-start
    
    #INPUTS:
    #qp         - dictionary which contains the optimization matrices
    #gen        - dictionary holding relevant information for all generation/load assets
    #market
    #subnet
    #options
    #forecast
    #first_profile  - dispatch from entire horizon optimization without lower bound constraints
    #scale_cost - list of scaling factors for fuel or electricity cost at ach time step
    #gen_output_w_starts - matrix of generator setpoints when timesteps are optimized individually
    #stor_power - power from storage components
    #binary_comb- list of identified feasible unit commitment combinations 
    #disp_comb  - Optimal setpoint for each unit commitment combination
    #cost_comb  -  cost for each combination
    #verified   - binary vector indicating if a combinations has been verified feasible
    #observer   - 
    #dt         - vector of length of each timestep
    #vh         - value heat from CHP generators
    

    #OUTPUTS:
    #gen_output - matrix of generator setpoints when timesteps are optimized individually, and startup cost is minimized
    #i_best     - 
    '''
    n_s = len(dt)
    n_g = len(gen)
    include = ['ElectricGenerator','CombinedHeatPower','Heater','Electrolyzer','HydrogenGenerator','CoolingTower','Chiller']
    skip_on = {'gen':[],'start':[],'stop':[],'duration':[]}
    skip_off = {'gen':[],'start':[],'stop':[],'duration':[]}
    locked = [[True for j in range(n_s+1)] for i in range(n_g)]
    max_start_cost = max([gen[k]['start_cost'] for k in range(n_g) if 'start_cost' in gen[k]])
    gen_output = [[i for i in j] for j in gen_output_w_starts]
    inc = [False for i in range(n_g)]
    for i in range(n_g):
        if gen[i]['type'] in include:
            locked[i] = [gen_output[i][t]>0 for t in range(n_s+1)]
            inc[i] = True
    if 'renewable' in forecast:
        for i in range(n_g):
            if len(forecast['renewable'][i])>0:
                for j in range(len(dt)):
                    gen_output[i][j+1] += forecast['renewable'][i][j]
    i_best = [min(range(len(cost_comb[t])), key=cost_comb[t].__getitem__)for t in range(n_s)] #find i_best(t), index that results in lowest cost once start-up is considered
    
    bat = [[j for j in observer['building_avg_temp']]]
    flt = [[j for j in observer['fluid_loop_temperature']]]
    flc = observer['fluid_loop_capacitance']
    
    bat.extend([build_temp[t][i_best[t]] for t in range(n_s)])
    flt.extend([fluid_temp[t][i_best[t]] for t in range(n_s)])

    seg = 1 #track the shortest ___ segments, currently set to 20. If fewer than 20 segments it ends early
    while seg < 20:
        #try and remove the shortest segment with generator on, and if there are multiple segments 
        #of equal length, start with highest startup cost.
        #if the segments are equal length and same generator, then start with the first segment
        on_seg, off_seg = seg_length(locked, gen, dt, skip_on, skip_off, inc)
        if on_seg['gen'] == [] and off_seg['gen'] == []:
            seg = float('inf')
        else:
            osl = len(on_seg['gen'])
            weighted_value = [on_seg['duration'][i]-gen[on_seg['gen'][i]]['start_cost']/(2*max_start_cost)+on_seg['start'][i]/n_s for i in range(osl)]
            weighted_value.extend([off_seg['duration'][i]-gen[off_seg['gen'][i]]['start_cost']/(2*max_start_cost)+off_seg['start'][i]/n_s for i in range(len(off_seg['gen']))])
            i_seg = min(range(len(weighted_value)), key=weighted_value.__getitem__)
            if i_seg<osl:
                k = on_seg['gen'][i_seg]
                t1 = on_seg['start'][i_seg]
                t2 = on_seg['stop'][i_seg]
                i_best, locked, no_alt, disp_comb, cost_comb, verified = alt_generation(
                    qp, gen, market, subnet, options, first_profile, stor_power, forecast, scale_cost,
                    i_best, k, t1, t2, binary_comb, disp_comb, cost_comb, verified, flt, flc, bat, gen_output, locked, dt, v_h)
                if no_alt:
                    #second, can it get close to the final storage capacity without this on-segment
                    binary_comb, disp_comb, cost_comb, verified, i_best, locked, must_replace = avoid_generation(
                        qp, gen, market, subnet, options, first_profile, stor_power, forecast, scale_cost,
                        i_best, k, t1, t2, binary_comb, disp_comb, cost_comb, verified, flt, flc, bat, gen_output, locked, dt, v_h)
                    if must_replace:
                        #third try moving generation earlier or later with same generator (if there is storage)
                        binary_comb, disp_comb, cost_comb, verified, i_best, locked, cant_move = move_generation(
                            qp, gen, market, subnet, options, first_profile, stor_power, forecast, scale_cost,
                            i_best, k, t1, t2, binary_comb, disp_comb, cost_comb, verified, flt, flc, bat, gen_output, locked, inc, dt, v_h) 
                        if cant_move:#add to list of segments to avoid
                            skip_on['gen'].append(on_seg['gen'][i_seg])
                            skip_on['start'].append(on_seg['start'][i_seg])
                            skip_on['stop'].append(on_seg['stop'][i_seg])
                            skip_on['duration'].append(on_seg['duration'][i_seg])
            else:
                i = i_seg - osl
                k = off_seg['gen'][i]
                t1 = off_seg['stop'][i]
                t2 = off_seg['start'][i]
                i_best, locked, cant_keep_on, disp_comb, cost_comb, verified = leave_gen_on(
                    qp, gen, market, subnet, options, first_profile, stor_power, forecast, scale_cost,
                     i_best, k, t1, t2, binary_comb, disp_comb, cost_comb, verified, flt, flc, bat, gen_output, locked, dt, v_h)
                if cant_keep_on:#add to list of segments to avoid
                    skip_off['gen'].append(off_seg['gen'][i])
                    skip_off['start'].append(off_seg['start'][i])
                    skip_off['stop'].append(off_seg['stop'][i])
                    skip_off['duration'].append(off_seg['duration'][i])
            update_gen_output(gen, gen_output, disp_comb, i_best, dt)
            seg = seg+1
    return gen_output, i_best


def update_gen_output(gen, gen_output, disp_comb, i_best, dt):
    #pull the corresponding best dispatches with the start-cost considered
    for t in range(len(dt)):
        status = [gen_output[i][t] for i in range(len(gen_output))]
        new_status = [j for j in disp_comb[t][i_best[t]]]
        for i in range(len(gen)):
            if gen[i]['type'] in ['ElectricStorage', 'ThermalStorage']:
                loss = gen[i]['stor']['self_discharge']*gen[i]['stor']['usable_size']
                if (new_status[i] +loss) > 0: #discharging
                    d_soc = -(new_status[i] + loss)*dt[t]/gen[i]['stor']['disch_eff']
                else: #charging
                    d_soc = -(new_status[i] + loss)*dt[t]*gen[i]['stor']['charge_eff']
                new_status[i] = max([min([status[i] + d_soc,gen[i]['stor']['usable_size']]),0])
        for i in range(len(gen_output)):
            gen_output[i][t+1] = new_status[i]


def avoid_generation(qp, gen, market, subnet, options, first_profile, stor_power, forecast, scale_cost, 
                    i_best, k, t1, t2, binary_comb, disp_comb, cost_comb, 
                    verified, flt, flc, bat, gen_output, locked, dt, v_h):
    #TODO validate for variable time steps
    '''
    Find the cheapest feasible alternative dispatch without this generator
    # only use generators that were on previously or will be on
    #look for enough slack in other generators to avoid the overdepleating
    #storage at the minimum and to get to the same final state
    '''
    from eagers.basic.component_output import eff_interp

    n_s = len(dt)
    must_replace = True
    cap = gen[k]['cap']
    eff = gen[k]['eff']
    stor_energy, _, stor, _, _ = stor_state(gen, gen_output, dt)
    out = None
    if gen[k]['type'] == 'Chiller':
        out = 'c'
    elif gen[k]['type'] == 'Heater':
        out = 'h'
    elif gen[k]['type'] in ('CombinedHeatPower', 'ElectricGenerator'):
        if 'dc' in stor:
            out = 'dc'
        elif 'e' in stor:
            out = 'e'
    #TODO add other dispatchable components
    if not out is None and out in stor and any([j>0 for j in stor_energy[out]]):
        _, _, spare_capacity, _ = gen_limit(gen, gen_output, locked, dt,[out])
        rmv_cost = 0
        margin_cost = gen_marginal_cost(gen, gen_output, scale_cost, dt)
        rem_gen = [0 for t in range(n_s)]
        rem_heat = [0 for t in range(n_s)]
        for t in range(t1,t2,1):
            #need to replace this energy in the storage by the end and 
            #replace enough early so that UsefulStoredEnergy - remStor + makeup does not go negative
            eff_t =eff_interp(cap,eff,gen_output[k][t+1])
            rem_gen[t:] = [rem_gen[tt] + gen_output[k][t+1]*dt[t] for tt in range(t,len(dt))]
            if gen[k]['type'] == 'CombinedHeatPower':
                rem_heat[t:] = [rem_heat[tt] + chp_heat(gen[k],[gen_output[k][t+1]])[0]*dt[t] for tt in range(t,len(dt))]
            rmv_cost += scale_cost[k][t+1]*gen_output[k][t+1]/eff_t*dt[t]
            if 'const_cost' in gen[k]:
                rmv_cost += gen[k]['const_cost'] * scale_cost[k][t+1] * dt[t]
            if gen[k]['type'] == 'Chiller' and 'electrical' in gen[k]['const_demand']:
                rmv_cost +=  (gen[k]['const_demand']['electrical'] + (gen_output[k][t+1]/eff_t))*min(margin_cost['e']['cost']['spin_reserve'][k][t])*dt[t]
        if gen[k]['type'] == 'CombinedHeatPower' and not 'h' in stor:
            stor_energy['h'] = [0 for j in range(n_s)]
            
        possible_to_avoid = True    
        if (sum(spare_capacity[out])+0.5*stor_energy[out][-1])<0: #it can reach at least half the planned final SOC
            possible_to_avoid = False
        for t in range(n_s):
            if (sum(spare_capacity[out][:t+1]) - sum(rem_gen[:t+1]))<-stor_energy[out][t]: #storage never lets the SOC dip below 0
                possible_to_avoid = False
            if gen[k]['type'] == 'CombinedHeatPower' and (sum(spare_capacity['h'][:t+1]) - sum(rem_heat[:t+1]))<-stor_energy['h'][t]:
                possible_to_avoid = False
        if possible_to_avoid:
            for t in range(t1,t2):
                new_comb = [locked[j][t+1] for j in range(len(locked))]
                new_comb[k] = False #turn off
                new_case = find_case(binary_comb[t], new_comb)#find indicies of all combinations that include locked[t+1,:]
                if not new_case is None:
                    if not verified[t][new_case]:
                        cost_comb[t], verified[t], disp_comb[t] = test_more_cases(
                            qp, [new_case], cost_comb[t], verified[t], disp_comb[t], stor_power[t], flt[t], flc, bat[t], binary_comb[t],
                            gen, market, subnet, options, first_profile, forecast, scale_cost, dt, t, any([v_h[n][t] for n in range(len(v_h))]))
                    if not cost_comb[t][new_case] == float('inf'):
                        i_best[t] = new_case
                    else:
                        possible_to_avoid = False
                else:
                    disp_comb[t].append([j for j in disp_comb[t][i_best[t]]])
                    binary_comb[t].append([j for j in binary_comb[t][i_best[t]]])
                    cost_comb[t].append(cost_comb[t][i_best[t]]-0.001) #make the other combos barely cheaper than the standard
                    verified[t].append(False)
                    i_best[t] = len(verified[t]) - 1 # -1 because it is an index
                    binary_comb[t][i_best[t]][k] = False #set output of this generator to zero
                    disp_comb[t][i_best[t]],exceeding_charge_limit = stor_add(gen, disp_comb[t][i_best[t]], -disp_comb[t][i_best[t]][k], k)
                    disp_comb[t][i_best[t]][k] = 0 #set output of this generator to zero
            if possible_to_avoid:
                must_replace = False
                for t in range(t1,t2):
                    locked[k][t+1] = False
        #try and replace the generation with the cheapest generation from any other spare capacity
        #otherwise storage at end of horizon drops
        if possible_to_avoid:
            #need to make up generation before storage would go negative
            t_limit = 0
            neg_stor = [j for j in range(n_s) if rem_gen[j]>stor_energy[out][j]]
            if len(neg_stor)>0:
                t_limit = neg_stor[0]
            else:
                t_limit = n_s
            make_up_gen = 0
            rem_e = max(rem_gen) #removed energy (maximum = sum of generator output that is turned off * dt)
            sorted_mc = sort_mc(margin_cost, out, t_limit, k, dt)
            if not sorted_mc is None and sorted_mc['cumulative_energy'][-1] >= rem_e:
                j = 0
                while sorted_mc['cumulative_energy'][j]<rem_e:
                    j+=1
                if j>0:
                    make_up_cost = sorted_mc['cumulative_cost'][j-1] + sorted_mc['marginal_cost'][j]*(rem_e - sorted_mc['cumulative_energy'][j-1])
                else:
                    make_up_cost = sorted_mc['marginal_cost'][0]*rem_e
                if (make_up_cost-rmv_cost)<gen[k]['start_cost']: #if spare_gen is less than start_cost
                    r=0
                    while make_up_gen<rem_e and r<len(sorted_mc['cumulative_energy'])-1:
                        t = int(sorted_mc['timestep'][r])
                        add_gen_energy = min([sorted_mc['capacity'][r]*dt[t],rem_e-make_up_gen])
                        disp_comb[t][i_best[t]][sorted_mc['generator'][r]] = disp_comb[t][i_best[t]][sorted_mc['generator'][r]] + add_gen_energy/dt[t]
                        #edit storage dispatch at this timestep so that update_storate_state works correctly
                        disp_comb[t][i_best[t]],exceeding_charge_limit = stor_add(gen, disp_comb[t][i_best[t]], add_gen_energy, k)
                        make_up_gen += (add_gen_energy - exceeding_charge_limit)
                        r+=1    
    return binary_comb, disp_comb, cost_comb, verified, i_best, locked, must_replace


def find_case(binary_comb,case):
    k = None
    for i in range(len(binary_comb)):
        if binary_comb[i] == case:
            k = i
            break
    return k


def move_generation(qp, gen, market, subnet, options, first_profile, stor_power,
    forecast,  scale_cost, i_best, k, t1, t2, binary_comb, 
    disp_comb, cost_comb, verified, flt, flc, bat, gen_output, locked, inc, dt, v_h):
    #TODO validate for variable timestep
    stor_energy, _, _, spare_stor_cap, _ = stor_state(gen, gen_output,dt)
    out = None
    if gen[k]['type'] in ['CombinedHeatPower','ElectricGenerator']:
        if 'dc' in stor_energy:
            out = 'dc'
        elif 'e' in stor_energy:
            out = 'e'
    elif gen[k]['type'] == 'Chiller':
        out = 'c'
    elif gen[k]['type'] == 'Heater':
        out = 'h'
    elif gen[k]['type'] == 'CoolingTower':
        out = 'cw'
    else: 
        out = None
    ##TODO add more dispatchable components

    n_s = len(disp_comb)
    cant_move = True
    i_best_alt = [j for j in i_best]

    if out in stor_energy and any([x>0 for x in stor_energy[out]]):
        add_stor = [0 for t in range(n_s)]
        rem_stor = [0 for t in range(n_s)]
        stops = [t for t in range(n_s) if not locked[k][t+1] and locked[k][t]]
        n= t2-t1 #number of steps generator is on for
        pre_stops = [j for j in stops if j<t1]
        if len(pre_stops)>0:
            #it was on previously, try to move earlier if storage permits
            t_stop = pre_stops[-1]
            #check if there is a feasible option to leave this generator on at earlier time steps
            for j in range(n):
                tj = t_stop+j
                opt = binary_comb[tj] #all feasible options tested at this time
                locked_now = opt[i_best_alt[tj]]
                locked_now[k] = True
                new_locked = find_case(opt,locked_now)
                if not new_locked is None:
                    if not verified[tj][new_locked]:
                        cost_comb[tj],verified[tj],disp_comb[tj] = test_more_cases(
                            qp,[new_locked],cost_comb[tj],verified[tj],disp_comb[tj], stor_power[tj], flt[tj], flc, bat[tj], opt,
                            gen, market, subnet, options, first_profile, forecast, scale_cost, dt, tj, any([v_h[n][tj] for n in range(len(v_h))]))
                        if not cost_comb[tj][new_locked] == float('inf'):
                            i_best_alt[tj] = new_locked
                for ts in range(t1-t_stop):
                    add_stor[t_stop+ts+j] += gen_output[k][t1+j+1]
            if not any([i_best_alt[ts] == i_best[ts] for ts in range(t_stop,t_stop+n)]) and all([add_stor[t]<spare_stor_cap[out][t] for t in range(n_s)]):
                cant_move = False
                for kg in inc:
                    for j in range(n):
                        locked[kg][t_stop+j+1] = binary_comb[t_stop+j][i_best_alt[t_stop+j]][kg]
                for t in range(t_stop+n,t2):
                    new_comb = [locked[j][t+1] for j in range(len(locked))]
                    new_comb[k] = False
                    new_case = find_case(opt,new_comb)
                    if not new_case is None:
                        if not verified[t][new_case]:
                            cost_comb[t],verified[t],disp_comb[t] = test_more_cases(
                                qp,[new_case],cost_comb[t],verified[t],disp_comb[t],stor_power[t], flt[t], flc, bat[t],binary_comb[t], 
                                gen, market, subnet, options, first_profile, forecast, scale_cost, dt, t, any([v_h[n][t] for n in range(len(v_h))]))
                        if not cost_comb[t][new_case] == float('inf'):
                            i_best_alt[t] = new_case
                        else:
                            cant_move = True
                    else:
                        disp_comb[t].append(disp_comb[t][i_best_alt[t]])
                        binary_comb[t].append(binary_comb[t][i_best_alt[t]])
                        cost_comb[t].append(cost_comb[t][i_best_alt[t]]-0.001) #make the other combos barely cheaper than the standard
                        verified[t].append(False)
                        i_best_alt[t] = len(verified[t]) - 1 # -1 because it is an index
                        disp_comb[t][i_best_alt[t]][k] = 0 #set output of this generator to zero
                        binary_comb[t][i_best_alt[t]][k] = False #set output of this generator to zero
                if not cant_move:
                    for t in range(t_stop+n,t2):
                        locked[k][t+1] = False
                    i_best = [j for j in i_best_alt]
                
        if cant_move:
            starts = [t for t in range(n_s) if locked[k][t+1] and not locked[k][t]]
            t_start = [j for j in starts if j>t2]
            if len(t_start) == 0:
                t_start = n_s -n
            else:
                t_start = t_start[0]-n
            for j in range(n):
                if t_start+j<n_s:
                    opt = binary_comb[t_start+j] #all feasible options tested at this time
                    locked_now = opt[i_best_alt[t_start+j]]
                    locked_now[k] = True
                    new_locked = find_case(opt,locked_now)
                    if not new_locked is None:
                        i_best_alt[t_start+j] = new_locked
                    #need to hold this shifted energy in the storage from the previous time the generator shut down until it had come on before     
                    for ts in range(t_start-t1-1):
                        rem_stor[t1+j+ts] += gen_output[k][t1+j+1]
            if not any([i_best_alt[ts] == i_best[ts] for ts in range(t_start,t_start+n)]) and all([rem_stor[t]>stor_energy[out][t] for t in range(n_s)]):
                i_best = [j for j in i_best_alt]
                for j in range(n):
                    for kg in inc:
                        locked[kg][t_start+j+1] = binary_comb[t_start+j][i_best[t_start+j]][inc]
                    disp_comb[t_start+j][i_best[t_start+j]][k] = gen_output[k][t1+j+1]
                for t in range(t1,t_start):
                    locked[k][t+1] = False
                    disp_comb[t][i_best[t]][k] = 0 #set output of this generator to zero
                cant_move = False

    return binary_comb, disp_comb, cost_comb, verified, i_best, locked, cant_move


def alt_generation(qp, gen, market, subnet, options, first_profile, stor_power,
    forecast, scale_cost, i_best, k, t1, t2, binary_comb, 
    disp_comb, cost_comb, verified, flt, flc, bat, gen_output, locked, dt, v_h):
    '''
    #determine any possible subsitute generators
    '''
    stor_energy, stor_gen_avail, _, _, _  = stor_state(gen, gen_output, dt)
    no_alt = True
    n_g = len(gen)
    inc = [False for i in range(n_g)]
    acdc = False
    for i in range(n_g):
        if gen[i]['type'] == 'ACDCConverter':
            acdc = True
            dc_ac = gen[i]['output']['e'][0][1]
            ac_dc = gen[i]['output']['dc'][0][0]
    if gen[k]['type'] == 'Chiller':
        out = 'c'
        for i in range(n_g):
            if gen[i]['type'] == 'Chiller' or (gen[i]['type'] == 'ThermalStorage' and gen[i]['source'] == 'cooling'):
                inc[i] = True
    if gen[k]['type'] == 'CoolingTower':
        out = 'cw'
        for i in range(n_g):
            if gen[i]['type'] == 'CoolingTower':
                inc[i] = True
    if gen[k]['type'] == 'Heater':
        out = 'h'
        for i in range(n_g):
            if gen[i]['type'] in ['Heater', 'CombinedHeatPower'] or (gen[i]['type'] == 'ThermalStorage' and gen[i]['source']=='heat'):
                inc[i] = True
    if gen[k]['type'] in ['CombinedHeatPower', 'ElectricGenerator']:
        if 'dc' in gen[k]['output']:
            out = 'dc'
        elif 'e' in gen[k]['output']:
            out = 'e'
        for i in range(n_g):
            if gen[i]['type'] in ['CombinedHeatPower', 'ElectricGenerator', 'ElectricStorage']:
                inc[i] = True
            if gen[k]['type'] == 'CombinedHeatPower' and (gen[i]['type'] == 'Heater' or (gen[i]['type'] == 'ThermalStorage' and gen[i]['source'] == 'heat')):
                inc[i] = True
    #TODO add other dispatchable components
    if not out in stor_energy:
        stor_energy[out] = [0 for t in range(len(dt))]
        stor_gen_avail[out] = [0 for t in range(len(dt))]
    if 'dc' in stor_energy and acdc and out == 'e':
        stor_energy['e'] = [stor_energy['e'][t] + stor_energy['dc'][t]*dc_ac for t in range(len(dt))]
        stor_gen_avail['e'] = [stor_gen_avail['e'][t] + stor_gen_avail['dc'][t]*dc_ac for t in range(len(dt))]
    elif 'e' in stor_energy and acdc and out == 'dc':
        stor_energy['dc'] = [stor_energy['e'][t] + stor_energy['dc'][t]*ac_dc for t in range(len(dt))]
        stor_gen_avail['dc'] = [stor_gen_avail['dc'][t] + stor_gen_avail['e'][t]*dc_ac for t in range(len(dt))]
    
    i_best_alt = [j for j in i_best]
    d_cost = [0 for t in range(t2-t1)]
    d_gen = [0 for t in range(t2-t1)]
    alt_locked = [[i for i in j] for j in locked]
    #for the time that the generator in question is on, try combinations that don't have it on
    for t in range(t1,t2,1):
        opt = binary_comb[t] #all options at this time
        #run all unverified cases that replace he generator in question with either 1 or 2 or 3 others that are:
        #a) running within (t2-t1) steps before/after and have a marginal cost less than starting the generator in question
        #b) have a marginal cost + start-up cost less than starting the generator in question
        need_test = []
        for c in range(len(opt)):
            if not verified[t][c] and not opt[c][k]:
                new_on = [i for i in range(n_g) if opt[c][i] and not opt[i_best[t]][i]]
                d_c = cost_comb[t][c] - cost_comb[t][i_best[t]]
                if len(new_on)<=3 and all([inc[j] for j in new_on]):
                    for j in range(len(new_on)):
                        if not any(locked[new_on[j]]):
                            d_c += gen[new_on[j]]['start_cost']
                    if d_c<gen[k]['start_cost']:
                        need_test.append(c)
        cost_comb[t], verified[t], disp_comb[t] = test_more_cases(
            qp, need_test, cost_comb[t], verified[t], disp_comb[t], stor_power[t], flt[t], flc, bat[t], opt,
            gen, market, subnet, options, first_profile, forecast,  scale_cost, dt, t, any([v_h[n][t] for n in range(len(v_h))]))

        #determine the marginal cost of switching to the best possible option that doesn't use the generator in question
        cost2 = []
        for c in range(len(cost_comb[t])):
            if opt[c][k] or not verified[t][c]:
                cost2.append(float('inf')) #make the cost infinite for all combinations with generator i or untested combinations
            else:
                cost2.append(cost_comb[t][c] - cost_comb[t][i_best[t]])#cost for these options
       
        frac_time = dt[t]/sum(dt[t1:t2])
        for j in range(n_g):
            if 'start_cost' in gen[j] and gen[j]['start_cost']>0:
                for c in range(len(cost2)):
                    if (not cost2[c] == float('inf') 
                        and (not any([locked[j][t] for t in range(t1,t2+2)]) and opt[c][j])
                        or (all([locked[j][t] for t in range(t1,t2+2)]) and not opt[c][j])):#would be adding a startup or a temporary shutdown
                        cost2[c] += gen[j]['start_cost']*frac_time 
        if not all([cost2[c] == float('inf') for c in range(len(cost2))]):
            i_best_alt[t] = min(range(len(cost2)), key=cost2.__getitem__)
            d_cost[t-t1] = cost2[i_best_alt[t]]
            d_gen[t-t1] = disp_comb[t][i_best[t]][k] - disp_comb[t][i_best_alt[t]][k]
            for i in range(n_g):
                alt_locked[i][t+1] = binary_comb[t][i_best_alt[t]][i]
        else:
            #no feasible combinations without turning on a different generator, so don't make changes to this segment
            d_cost = [float('inf')]
            break
    #if it wants to swap to have a different generator on now, don't constrain its maximum output by shutting it down immediately
    for i in range(n_g):
        if all([binary_comb[t][i_best_alt[t]][i] and not binary_comb[t][i_best[t]][i] for t in range(t1,t2,1)]) and gen[i]['start_cost']>0:
            alt_locked[i][t2:] = [True for j in range(len(alt_locked[i])-t2)]
    #check if the alternate operation strategy is feasible when put in sequence
    _,_,_,spare_gen = gen_limit(gen,gen_output,alt_locked,dt, [out])
    spare_storage = spare_storage_check(spare_gen[out][t1:t2],stor_gen_avail[out][t1:t2],stor_energy[out][t1:t2],dt[t1:t2])
    if sum(d_cost)<gen[k]['start_cost'] and spare_storage:
        spare_heat_storage = True
        if gen[k]['type'] == 'CombinedHeatPower' and 'h' in stor_energy:
            spare_heat_storage = spare_storage_check(spare_gen['h'][t1:t2],stor_gen_avail['h'][t1:t2],stor_energy['h'][t1:t2],dt[t1:t2])
        if spare_heat_storage:
            i_best = [j for j in i_best_alt] #use the alternative index
            for t in range(t1,t2,1):
                for i in range(n_g):
                    locked[i][t+1] = binary_comb[t][i_best[t]][i]
            no_alt = False
    return i_best, locked, no_alt, disp_comb, cost_comb, verified


def spare_storage_check(spare_gen,stor_gen,stored_energy,dt):
    spare_storage = True
    if any([(spare_gen[j]+stor_gen[j])<0 for j in range(len(spare_gen))]):
        spare_storage = False #there is insufficient spare capacity in the other generators & storage at any moment
    else:
        for i in range(len(spare_gen)):
            if sum([spare_gen[j]*dt[j] for j in range(i+1)])+stored_energy[i]<0:
                spare_storage = False #the cumulative loss of generation would deplete the storage
    return spare_storage


def leave_gen_on(qp, gen, market, subnet, options, first_profile, stor_power,
    forecast, scale_cost, i_best, k, t1, t2, binary_comb,
     disp_comb, cost_comb,  verified, flt, flc, bat, gen_output, locked, dt, v_h):
    '''
    find the cheapest feasible alternative dispatch that keeps this generator on
    only use generators that were on previously or will be on
    '''
    n_g = len(gen)
    _, _, _, spare_stor_cap, stor_slack_avail = stor_state(gen, gen_output, dt)
    out = None
    if gen[k]['type'] in ['CombinedHeatPower', 'ElectricGenerator']:
        if 'dc' in spare_stor_cap:
            out = 'dc'
        elif 'e' in spare_stor_cap:
            out = 'e'
        else:
            out = None
    elif gen[k]['type'] == 'Chiller':
        out = 'c'
    elif gen[k]['type'] == 'Heater':
        out = 'h'
    elif gen[k]['type'] == 'CoolingTower':
        out = 'cw'
    turndown = turndown_capacity(gen,gen_output,k)
    cant_keep_on = True
    #only allow generators that are on at beginning or end to be involved or that have smaller start-up cost
    i_best_alt = [j for j in i_best]
    has_changed = False
    d_cost = [0 for t in range(t2-t1)]
    d_gen = [0 for t in range(t2-t1)]
    slack_gen = [0 for t in range(t2-t1)]
    if t1 == 0:
        prev = [gen_output[i][0] for i in range(n_g)]
    else:
        prev = disp_comb[t1-1][i_best[t1-1]]
    for t in range(t1,t2,1):
        opt = binary_comb[t] #all feasible options tested at this time
        #run unverified cases that just add the generator in question
        locked_now = opt[i_best_alt[t]]
        locked_now[k] = True
        new_locked = find_case(opt,locked_now)
        if not new_locked is None:
            if not verified[t][new_locked]:
                cost_comb[t], verified[t], disp_comb[t] = test_more_cases(
                    qp, [new_locked], cost_comb[t], verified[t], disp_comb[t], stor_power[t], flt[t], flc, bat[t], opt,
                    gen, market, subnet, options, first_profile, forecast, scale_cost, dt, t, any([v_h[n][t] for n in range(len(v_h))]))
            if not cost_comb[t][new_locked] == float('inf'):
                i_best_alt[t] = new_locked
                has_changed = True
                d_cost[t-t1] = cost_comb[t][i_best_alt[t]]-cost_comb[t][i_best[t]]
                d_gen[t-t1], slack_gen[t-t1] = slack_cap(gen, disp_comb[t][i_best[t]], disp_comb[t][i_best_alt[t]], prev, k, sum(dt[t1:t+1]))
        else:
            #no feasible combinations keeping this generator active
            #don't make changes to this segment
            d_cost = [float('inf')]
            break
    if has_changed:
        useful1 = [0 for t in range(t2-t1)]
        useful2 = 0
        if out in spare_stor_cap:
            useful1 = stor_slack_avail[out][t1:t2]
            useful2 = min(spare_stor_cap[out][t1:])
        #sum of the marginal increase in cost is less than the start-up cost, there is spare capacity in the other generators and storage
        #and the cumulative loss of generation does not deplete the storage
        
        if sum(d_cost)<gen[k]['start_cost'] and all([d_gen[t]<(slack_gen[t]+useful1[t]/dt[t1+t]) for t in range(len(d_gen))]) and (sum(d_gen)-sum(slack_gen))<useful2+sum(turndown):
            i_best = [j for j in i_best_alt]
            for t in range(t1, t2, 1):
                #best alternative option tested at this time
                for i in range(n_g):
                    locked[i][t+1] = binary_comb[t][i_best[t]][i]
            cant_keep_on = False
    return i_best, locked, cant_keep_on, disp_comb, cost_comb, verified


def seg_length(locked, gen, dt, skip_on, skip_off, inc):
    on_seg = {'gen':[],'start':[],'stop':[],'duration':[]}
    off_seg = {'gen':[],'start':[],'stop':[],'duration':[]}
    def check_seg_list(gen_i,stop,start,skip_off):
        same = False
        for i in range(len(skip_off['gen'])):
            if skip_off['gen'][i] == gen_i and skip_off['stop'][i] == stop and skip_off['start'][i] == start:
                same = True
                break
        return same

    n_g = len(gen)
    n_s = len(dt)
     #find length of segments that a generator is on or off
    for i in range(n_g):
        if inc[i] and gen[i]['start_cost']>0 and sum(locked[i])<len(locked[i]):
            starts = [t for t in range(n_s) if locked[i][t+1] and not locked[i][t]]
            stops = [t for t in range(n_s) if not locked[i][t+1] and locked[i][t]]
            n_on = len(starts)
            n_off = len(stops)
            #only look at generators that turn both on and off during the window
            if n_on>0 and n_off>0:
                #generator is off for a segment
                if stops[0]<starts[0]:
                    if not check_seg_list(i,stops[0],starts[0],skip_off): 
                        off_seg['gen'].append(i)
                        off_seg['stop'].append(stops[0])
                        off_seg['start'].append(starts[0])
                        off_seg['duration'].append(sum(dt[stops[0]:starts[0]+1]))
                    del stops[0]
                    n_off -= 1
                j = 0
                while j<n_off:
                    #index of generator, start index, stop index, duration of segment
                    if not check_seg_list(i,stops[j],starts[j],skip_on): 
                        on_seg['gen'].append(i)
                        on_seg['start'].append(starts[j])
                        on_seg['stop'].append(stops[j])
                        on_seg['duration'].append(sum(dt[starts[j]:stops[j]+1]))
                    j+=1
                    if j<n_on: 
                        if not check_seg_list(i,stops[j-1],starts[j],skip_off): 
                            off_seg['gen'].append(i)
                            off_seg['stop'].append(stops[j-1])
                            off_seg['start'].append(starts[j])
                            off_seg['duration'].append(sum(dt[stops[j-1]:starts[j]+1]))
    return on_seg, off_seg


def slack_cap(gen, original, new, prev, k, dt):
    n_g = len(gen)
    d_gen = gen[k]['a']['lb'][-1]
    if gen[k]['type'] == 'Chiller':
        include = 'Chiller'
    elif gen[k]['type'] == 'Heater':
        include = 'Heater'
    elif gen[k]['type'] in ['CombinedHeatPower', 'ElectricGenerator']:
        include = ['CombinedHeatPower','ElectricGenerator']
    slack_gen = 0
    for i in range(n_g):
        if i!=k and gen[i]['type'] in include:
            if original[i]>0 and new[i]>0:
                slack_gen += min([original[i]-gen[i]['a']['lb'][-1], original[i]-(prev[i]-gen[i]['ramp']['b'][1]*dt)])# capacity to ramp down from the original set point is constrained by the ramp rate from the previous setpoint and the lower bound
            elif original[i]>0:
                slack_gen += original[i]
            elif new[i]>0:
                slack_gen -= new[i]
    return d_gen, slack_gen


def stor_state(gen, gen_output, dt):
    n_g = len(gen)
    n_s = len(dt)
    stored_energy = {}
    stor_gen_avail = {}
    spare_stor_cap = {}
    stor_slack_avail = {}
    stor = {}
    for i in range(n_g):
        if 'stor' in gen[i] and not gen[i]['type'] == 'HydroStorage':
            outs = list(gen[i]['output'].keys()) #find outputs of storage from possible outs
            for out in outs:
                if not out in stored_energy:
                    stored_energy[out] = [0 for t in range(n_s)]
                    stor_gen_avail[out] = [0 for t in range(n_s)]
                    spare_stor_cap[out] = [0 for t in range(n_s)]
                    stor_slack_avail[out] = [0 for t in range(n_s)]
                    stor[out] = []
                buff = 0
                if 'buffer' in gen[i]:
                    buff = gen[i]['stor']['usable_size']*(gen[i]['buffer']/100)
                for t in range(n_s):
                    stored_energy[out][t] += (gen_output[i][t+1]-buff)*gen[i]['stor']['disch_eff']
                    stor_gen_avail[out][t] += min([(gen_output[i][t+1]-buff)*gen[i]['stor']['disch_eff']/dt[t], gen[i]['stor']['peak_disch']])
                    spare_stor_cap[out][t] += (gen[i]['stor']['usable_size']-buff-gen_output[i][t+1]/gen[i]['stor']['charge_eff'])
                    stor_slack_avail[out][t] += min([(gen[i]['stor']['usable_size']-buff-gen_output[i][t+1])/(gen[i]['stor']['charge_eff']*dt[t]), gen[i]['stor']['peak_charge']])
                stor[out].append(i)
    return stored_energy, stor_gen_avail, stor, spare_stor_cap, stor_slack_avail


def turndown_capacity(gen,gen_output,k):
    n_g = len(gen)
    n_s = len(gen_output[0])-1
    turndown = [0 for i in range(n_s)]
    ac_dc = False
    for i in range(n_g):
        if gen[i]['type'] == 'ACDCConverter':
            ac_dc = True
    for i in range(n_g):
        if gen[k]['type'] in ['CombinedHeatPower','ElectricGenerator'] and gen[i]['type'] in ['CombinedHeatPower','ElectricGenerator']:
            out = gen[i]['output']
            if ac_dc or  ('dc' in out and 'dc' in gen[k]['output']) or ('e' in out and 'e' in gen[k]['output']):
                turndown = turndown_now(gen[i],gen_output[i],turndown) 
        elif gen[k]['type'] == 'Chiller' and gen[i]['type'] == 'Chiller':
            turndown = turndown_now(gen[i],gen_output[i],turndown) 
        elif gen[k]['type'] == 'Heater':
            if gen[i]['type'] == 'CombinedHeatPower':
                lb = turndown_now(gen[i],gen_output[i],'lb')  
                set_heat = chp_heat(gen[i],gen_output[i][1:])
                min_heat = chp_heat(gen[i],lb)
                turndown = [turndown[j] + max([0,set_heat[j] - min_heat[j]]) for j in range(n_s)]
            elif gen[i]['type'] == 'Heater':
                turndown = turndown_now(gen[i],gen_output[i],turndown)
        elif gen[k]['type'] == 'CoolingTower' and gen[i]['type'] == 'CoolingTower':
            turndown = turndown_now(gen[i],gen_output[i],turndown)  
    return turndown


def turndown_now(gen,output,turndown):
    n_s = len(output) - 1
    lb = []
    for t in range(n_s-1):
        s = gen['states'][1][0]
        lb.append(max([gen[s]['lb'][-1],output[t]-gen['ramp']['b'][1],output[t+2]-gen['ramp']['b'][1]]))
    lb.append(max([gen[s]['lb'][-1],output[t]-gen['ramp']['b'][1]]))
    if isinstance(turndown,list):
        turndown = [turndown[i] + max([0,output[i+1]-lb[i]]) for i in range(n_s)]
        return turndown
    else:
        return lb
        

def stor_add(gen, prev, add_gen, k):
    def stor_add2(gen,prev,add_gen):
        change = max(-gen['peak_disch']+prev, min(add_gen, gen['peak_charge']+prev))
        add_gen -= change
        dispatch = prev-change
        return dispatch,add_gen
    n_g = len(gen)
    dispatch = [j for j in prev]
    for i in range(n_g):
        if gen[k]['type'] == 'Chiller' and gen[i]['type'] == 'ThermalStorage' and gen[i]['source'] == 'cooling':
            dispatch[i],add_gen = stor_add2(gen[i]['stor'],prev[i],add_gen)
        if gen[k]['type'] == 'Heater' and gen[i]['type'] == 'ThermalStorage' and gen[i]['source'] == 'heat':
            dispatch[i],add_gen = stor_add2(gen[i]['stor'],prev[i],add_gen)
        if gen[k]['type'] in ['CombinedHeatPower', 'ElectricGenerator'] and gen[i]['type'] == 'ElectricStorage':
            dispatch[i],add_gen = stor_add2(gen[i]['stor'],prev[i],add_gen)
    return dispatch,add_gen


def test_more_cases(qp, need_test, cost_comb, verified, disp_comb, stor_power, flt, flc, bat, opt, gen, market, subnet, options, first_profile, forecast, scale_cost, dt, t, v_h):
    parallel = False
    if not need_test == []:
        n_g = len(gen)
        demand = all_demands(forecast, subnet,[t])
        net_abbrev = [subnet[net]['abbreviation'] for net in subnet['network_names']]
        step_profile = [first_profile[i][t+1] for i in range(n_g)]
        sc = [[j[t]] if len(j)>0 else [] for j in scale_cost]
        marginal = net_marginal_cost(gen,net_abbrev,market,step_profile,sc,dt[t],v_h)#update marginal cost
        fb = building_forecast_now(forecast['building'],t)
        renewable = {}
        for j in range(len(gen)):
            if gen[j]['type'] in ['Solar','Wind']:
                renewable[gen[j]['name']] = forecast['renewable'][j][t]
        min_power, max_power = constrain_min_max(gen, 'initially constrained', [first_profile[i][0] for i in range(n_g)], dt[t], sum(dt[0:t+1]))
        update_matrices_step(gen,market,subnet,options,qp,
            fb,renewable,demand,sc,marginal,stor_power,
            dt[t],min_power, max_power,step_profile,flt,flc,bat)

        if parallel:
            pass
            # cost_par= [cost_comb[n] for n in need_test]
            # disp_comb_par = [disp_comb[n] for n in need_test]
            # opt_par = [opt[n] for n in need_test]
            # qp_par = qp
            # with multiprocessing.Pool() as pool:
            #     #variables in n_test long list, acting as parfor in matlab
            #     a = pool.starmap(mp_test_more_cases,zip(cost_par,disp_comb_par, opt_par, qp_par))
            # unzip_a = zip(*a)
            # for n in need_test:
            #     cost_comb[n] = next(unzip_a)
            #     verified[n] = True
            #     disp_comb[n] = next(unzip_a)
        else:
            for c in need_test:
                disable_gen(qp, opt[c])
                x, flag1 = ecos_qp(qp['h'], qp['f'], qp['a'], qp['b'], qp['a_eq'], qp['b_eq'], qp['ub'], qp['lb'], qp['x_keep'], qp['r_keep'], qp['req_keep'])
                if flag1 == 0:
                    cost_comb[c] = sum([0.5 * x[i] * qp['h'][i] * x[i] + x[i] * qp['f'][i] for i in range(len(x))]) + sum([qp['const_cost'][j] * opt[c][j] for j in range(len(opt[c]))])
                    verified[c] = True
                    disp_comb[c] = sort_gen_disp_step(x, qp)
                else:
                    cost_comb[c] = float('inf')
    return cost_comb, verified, disp_comb


def mp_test_more_cases(cost_par, disp_comb_par, opt_par, qp):
    # Multiprocessing for test_more_cases
    disable_gen(qp, opt_par)
    x_par, flag1 = ecos_qp(qp['h'], qp['f'], qp['a'], qp['b'], qp['a_eq'], qp['b_eq'], qp['ub'], qp['lb'], qp['x_keep'], qp['r_keep'], qp['req_keep'])
    n_g = len(qp['const_cost'])
    if flag1 == 0:
        cost_par = sum([0.5*x_par[i]*qp['h'][i]*x_par[i] + x_par[i]*qp['f'] for i in range(len(x_par))]) + sum([qp['const_cost'][j] * opt_par[j] for j in range(n_g)])
        disp_comb_par = sort_gen_disp_step(x_par, qp)
    else:
        cost_par = float('inf')
    return [cost_par, disp_comb_par]



'''
Interpolate test_data at the dates requested
Returns a list of dictionaries: each item in 
list is the data at one specific time
'''

def get_data(test_data,date):
    d_key = list(test_data.keys())
    req_data = []
    data = {}
    d_key.remove('date')
    for k in d_key:
        if type(test_data[k]) is dict:
            data[k] = {}
    n = len(test_data['date'])
    t = 0
    for d in date:
        while t<=n and test_data['date'][t+1]<d:
            t +=1
        if d<test_data['date'][0]:
            for k in d_key:
                if type(test_data[k]) is dict:
                    d2_key = list(test_data[k].keys())
                    for j in d2_key:
                        data[k][j] = test_data[k][j][0]
                else:
                    data[k] = test_data[k][0]
        elif d>test_data['date'][-1]:
            for k in d_key:
                if type(test_data[k]) is dict:
                    d2_key = list(test_data[k].keys())
                    for j in d2_key:
                        data[k][j] = test_data[k][j][-1]
                else:
                    data[k] = test_data[k][-1]
        else:
            r = (test_data['date'][t+1]-d)/(test_data['date'][t+1]-test_data['date'][t])
            for k in d_key:
                if type(test_data[k]) is dict:
                    d2_key = list(test_data[k].keys())
                    for j in d2_key:
                        if type(test_data[k][j][t]) is list:
                            data[k][j] = [r*test_data[k][j][t][i] + (1-r)*test_data[k][j][t+1][i] for i in range(len(test_data[k][j][t]))]
                        else:
                            data[k][j] = r*test_data[k][j][t] + (1-r)*test_data[k][j][t+1]
                else:
                    if type(test_data[k][t]) is list:
                        data[k] = [r*test_data[k][t][i] + (1-r)*test_data[k][t+1][i] for i in range(len(test_data[k][t]))]
                    else:
                        data[k] = r*test_data[k][t] + (1-r)*test_data[k][t+1]
        req_data.append(data)
    return req_data


def surface_forecast(prev_data, hist_prof, date, temperature,holidays):
    """Surface forecast."""
    forecast = 'forecast'

    # d_sim = prev.Timestamp(end);
    # days = ceil(date(end)-floor(d_sim)); %this says that you may need to use surface fits for multiple days.
    # s12hour = nnz(date<(d_sim+.5)); %steps in next 12 hours
    # a = datevec(d_sim);
    # date_days = floor(d_sim)-1:ceil(d_sim+days);
    # mon_fri = zeros(length(date_days),1);
    # for i = 1:1:length(date_days)
    #     if weekday(date_days(i))<=6 && weekday(date_days(i))>=2 && (isempty(holidays) || ~any(holidays==date_days(i)))
    #         mon_fri(i) = 1;
    #     else
    #         mon_fri(i) = 0;
    #     end
    # end
    # month_name = {'Jan' 'Feb' 'Mar' 'Apr' 'May' 'Jun' 'Jul' 'Aug' 'Sep' 'Oct' 'Nov' 'Dec' 'Jan'};
    # hour = (date-floor(d_sim))*24;
    # f = fieldnames(prev.Demand);
    # for s = 1:1:length(f) %repeat for electric, cooling, heating, and steam as necessary
    #     [~,n_d] = size(prev.Demand.(f{s}));
    #     demand.(f{s}) = zeros(length(date),n_d);
    #     list = fieldnames(hist_prof.(f{s})(1));
    #     for k = 1:1:n_d
    #         month = a(2);
    #         hist_fit_dem = nan(length(date),1);
    #         for j = 0:1:ceil(days)
    #             if a(3)+ (j-1) > (datenum(a(1), a(2)+1, 1)-datenum(a(1), a(2), 1))
    #                 month = a(2)+1;
    #             end
    #             %load suface
    #             if length(list)>1
    #                 if nnz(strcmp(list,strcat(month_name(month),'WeekEnd')))>0 %historical profile is split to weekday/weekend
    #                     if mon_fri(j+1) == 0
    #                         surface = hist_prof.(f{s})(k).(strcat(month_name{month},'WeekEnd'));
    #                     else
    #                         surface = hist_prof.(f{s})(k).(strcat(month_name{month},'WeekDay'));
    #                     end
    #                 elseif nnz(strcmp(list,char(month_name(month))))>0 %historical profile is broken by month, but not weekend/weekday
    #                     surface = hist_prof.(f{s})(k).(month_name{month});
    #                 else 
    #                     surface = hist_prof.(f{s})(k).(list{1});
    #                 end
    #             else
    #                 surface = hist_prof.(f{s})(k).(list{1});% only 1 historical surface fit
    #             end
    #             if j ==0 %surface for yesterday
    #                 yest_pred = surface(mod((prev.Timestamp-floor(prev.Timestamp(1)))*24,24),prev.Weather.DrybulbC);
    #                 error_pred = min(1,max(-1,(prev.Demand.(f{s})(:,k)- yest_pred)./yest_pred));
    #                 bias_perc = mean(error_pred)*100; %average percent error between hisorical expectations for yesterday, and actual yesterday
    #                 hourly_error = prev.Demand.(f{s})(:,k) - yest_pred*(1+0.7*bias_perc/100);%hourly variation with equal mean (historical and last 24 hours)
    #             else %surface for current day or subsequent days
    #                 index = (hour<=(24*j)) & (hour>=(24*(j-1))); %index of Time corresponding to this day
    #                 hist_fit_dem(index) = surface(mod(hour(index),24),temperature_forecast(index));
    #             end
    #         end
    #         offset = prev.Demand.(f{s})(end,k) - hist_fit_dem(1)*(1+0.7*bias_perc/100);
    #         demand.(f{s})(1:s12hour,k) = max(0,hist_fit_dem(1:s12hour)*(1+0.7*bias_perc/100) + linspace(.5,0,s12hour)'.*(interp1(prev.Timestamp,hourly_error,date(1:s12hour)-1) + .7*offset));
    #         demand.(f{s})(s12hour+1:end,k) = max(0,hist_fit_dem(s12hour+1:end)*(1+0.7*bias_perc/100));
    #         demand.(f{s})(abs(demand.(f{s}))<1e-2) = 0
    return forecast



from .basic.file_handling import ensure_dirpath
from .config.path_spec import AUTO_CREATED_USER_DIRS


# Ensure user directories exist whenever eagers is imported.
for user_dirpath in AUTO_CREATED_USER_DIRS:
    ensure_dirpath(user_dirpath)



"""Datetime extensions.

Functions:
datetime_from_data
round_datetime
timedelta_split
"""

import datetime as dt

from dateutil import parser

from eagers.config.precision import \
    DATETIME_ROUNDING_PRECISION, DATETIME_ROUNDING_TIMEDELTA_FIELD as DRTF


def datetime_from_data(d):
    """Converts string values to datetime and rounds datetime values."""
    if isinstance(d, dt.datetime):
        return round_datetime(d)
    elif isinstance(d, str):
        return parser.parse(d)
    else:
        raise TypeError(
            f"Unrecognized type for datetime data: {type(d).__name__!r}")


def round_datetime(d, precision=DATETIME_ROUNDING_PRECISION):
    """Round a datetime to the named precision. If no precision is
    specified, the default datetime precision will be used.

    Cf. https://stackoverflow.com/a/3464000/7232335
    """
    # Add half the precision.
    d += dt.timedelta(**{f"{precision}s": 1}) / 2
    # "Floor" using next highest precision.
    d -= dt.timedelta(**{f"{DRTF}s": getattr(d, DRTF)})
    return d


def timedelta_split(tdelta):
    """Returns a list of length 4 with number of days, hours, minutes,
    and seconds in the given timedelta.
    """
    hours, seconds = divmod(tdelta.seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    return [tdelta.days, hours, minutes, seconds + tdelta.microseconds / 1e6]



'''
Calculates the surface temperature of each window pane 
all temperatures in Kelvin for this function
T is a matrix where each row holds the temperature states for a single window. 
The first column is the inside surface of the innermost pane, second column is outside of first pane, third is inside of 2nd pane...
'''
from eagers.building.basic.natural_convection import natural_convection
from eagers.building.basic.window_convection import window_convection

import numpy as np

def window_temperature(building,wind_speed,leward,S,T,T_int,T_sky,T_i,T_o,T_s,cp):
    sig = 5.67e-8 # W/m^2*K^4 Stephan-Boltzman const
    windows = building.windows
    n_sur = len(building.surfaces['name'])
    n_win = len(windows['type'])
    T_new = [[0,0,0,0] for i in range(n_win)]
    k_air = [26.3e-3*windows['gap_thickness'][i] for i in range(n_win)]

    Z = [sig*windows['emittance'][i][1]*windows['emittance'][i][2]/(1-(1-windows['emittance'][i][1])*(1-windows['emittance'][i][2])) for i in range(n_win)]
    F_sky = [0.5*(1+windows['cos_phi'][i]) for i in range(n_win)]
    Eo = [sig*((0.5*(1-windows['cos_phi'][i]))*T_o[i]**4 + F_sky[i]**1.5*T_sky**4 + F_sky[i]*(1-F_sky[i]**.5)*T_o[i]**4) for i in range(n_win)]
    Ei_s = [sig*sum([building.zones['view_factors'][i][j+n_sur]*building.surfaces['area_rad'][i]*(T_s[i])**4 for i in range(n_sur)])/building.windows['area'][j] for j in range(n_win)] #internal radiation from other surfaces to window
    vf_w = [[building.zones['view_factors'][i+n_sur][j+n_sur] for j in range(n_win)] for i in range(n_win)] #window 2 window view factors
    error = 1
    count = 0
    r = .4
    while error>0.1*building.site['temp_tol'] and count<10:
        h_n_ext = natural_convection(windows['normal'], [T[i][0] - T_o[i] for i in range(n_win)],None,building.convection['exterior'])
        h_o = [0 for i in range(n_win)]
        if building.convection['exterior'] =='DOE-2':
            for i in range(n_win):
                if leward[i]:
                    h_o[i] = (h_n_ext[i]**2 + (3.55*wind_speed[i]**.617)**2)**.5
                else:
                    h_o[i] = (h_n_ext[i]**2 + ((3.26*wind_speed[i])**(.89))**2)**.5 
        elif building.convection['exterior'] == 'Detailed' or building.convection['exterior'] == 'BLAST' or building.convection['exterior'] == 'TARP':
           for i in range(n_win):
                if leward[i]:
                    h_o[i] = h_n_ext[i] + 2.537*.5*(windows['perimeter'][i]*wind_speed[i]/windows['area'][i])**.5
                else:
                    h_o[i] = h_n_ext[i] + 2.537*(windows['perimeter'][i]*wind_speed[i]/windows['area'][i])**.5
        h_i = window_convection(windows,T_int,T_i,cp)
        error = 0
        for i in range(n_win):
            Ei = Ei_s[i]+ sig*(sum([vf_w[i][j]*T_int[j]**4 for j in range(n_win)]))
            if windows['type'][i]=='SimpleGlazingSystem':
                k = windows['thermal_conductivity'][i]/windows['thickness'][i]
                h_r = [windows['emittance'][i][j]*sig*T[i][j]**3 for j in range(2)]
                e = windows['emittance'][i][0:2]
                A = np.array([[-h_r[0]-k-h_o[i],       k           ],
                              [      k         ,   -h_r[1]-k-h_i[i]]])
                B = np.array([-e[0]*Eo[i] - h_o[i]*T_o[i] - S[0][i],
                    -e[1]*Ei - h_i[i]*T_i[i] - S[1][i]])
                ans = np.linalg.solve(A,B)
            elif windows['type'][i]=='Glazing':
                k = [windows['thermal_conductivity'][i][j]/windows['thickness'][i][j] for j in range(len(windows['thermal_conductivity'][i]))]
                h_r = [windows['emittance'][i][j]*sig*T[i][j]**3 for j in range(4)]
                e = [windows['emittance'][i][0],windows['emittance'][i][3]]
                h_r[1] = Z[i]*T[i][1]**3
                h_r[2] = Z[i]*T[i][2]**3
                h1 = k_air[i]
                A = np.array([[-h_r[0]-k[0]-h_o[i],       k[0]       ,        0       ,      0             ],
                                [   k[0]            ,   -h_r[1]-k[0]-h1,      h1+h_r[2] ,      0             ],
                                [       0           ,       h1+h_r[1]  , -h1-k[1]-h_r[2],     k[1]           ],
                                [        0          ,         0        ,       k[1]     , -h_r[3]-k[1]-h_i[i]]]) 
                B = np.array([-e[0]*Eo[i] - h_o[i]*T_o[i] - S[0][i],
                                -S[1][i],
                                -S[2][i],
                    -e[1]*Ei - h_i[i]*T_i[i] - S[3][i]])
                ans = np.linalg.solve(A,B)
            T_new[i] = [float(ans[j]) for j in range(len(ans))]
            error = float(max(error,np.amax([abs(T_new[i][j] - T[i][j]) for j in range(len(T_new[i]))])))
       
        for i in range(n_win):
            T[i] = [(1-r)*T_new[i][j] + r*T[i][j] for j in range(len(T_new[i]))]
            if windows['type'][i] =='SimpleGlazingSystem':
                T_int[i] = T[i][1]
            else:
                T_int[i] = T[i][3]
        count +=1
    Q_windows = [0 for i in range(len(building.zones['name']))]
    for i in range(n_win):
        Q_windows[windows['zone_index'][i]] += h_i[i]*windows['area'][i]*(T_int[i] - T_i[i])
    T_windows_int = [T_int[i] - 273 for i in range(n_win)] #convert back to Celcius
    return T,T_windows_int,Q_windows,h_i


"""Implementation of LOESS smoothing from:
https://github.com/joaofig/pyloess

Functions:
tricubic
main

Classes:
Loess
"""

import numpy as np
import time
import math


def tricubic(x):
    y = np.zeros_like(x)
    idx = (x >= -1) & (x <= 1)
    y[idx] = np.power(1.0 - np.power(np.abs(x[idx]), 3), 3)
    return y


class Loess:

    def __init__(self, xx, yy, degree=1):
        self.n_xx, self.min_xx, self.max_xx = self.normalize_array(xx)
        self.n_yy, self.min_yy, self.max_yy = self.normalize_array(yy)
        self.degree = degree

    @staticmethod
    def normalize_array(array):
        min_val = np.min(array)
        max_val = np.max(array)
        return (array - min_val) / (max_val - min_val), min_val, max_val

    @staticmethod
    def get_min_range(distances, window):
        min_idx = np.argmin(distances)
        n = len(distances)
        if min_idx == 0:
            return np.arange(0, window)
        if min_idx == n-1:
            return np.arange(n - window, n)

        min_range = [min_idx]
        while len(min_range) < window:
            i0 = min_range[0]
            i1 = min_range[-1]
            if i0 == 0:
                min_range.append(i1 + 1)
            elif i1 == n-1:
                min_range.insert(0, i0 - 1)
            elif distances[i0-1] < distances[i1+1]:
                min_range.insert(0, i0 - 1)
            else:
                min_range.append(i1 + 1)
        return np.array(min_range)

    @staticmethod
    def get_weights(distances, min_range):
        max_distance = np.max(distances[min_range])
        weights = tricubic(distances[min_range] / max_distance)
        return weights

    def normalize_x(self, value):
        return (value - self.min_xx) / (self.max_xx - self.min_xx)

    def denormalize_y(self, value):
        return value * (self.max_yy - self.min_yy) + self.min_yy

    def estimate(self, x, window, use_matrix=False, degree=1):
        n_x = self.normalize_x(x)
        distances = np.abs(self.n_xx - n_x)
        min_range = self.get_min_range(distances, window)
        weights = self.get_weights(distances, min_range)

        if use_matrix or degree > 1:
            wm = np.multiply(np.eye(window), weights)
            xm = np.ones((window, degree + 1))

            xp = np.array([[math.pow(n_x, p)] for p in range(degree + 1)])
            for i in range(1, degree + 1):
                xm[:, i] = np.power(self.n_xx[min_range], i)

            ym = self.n_yy[min_range]
            xmt_wm = np.transpose(xm) @ wm
            beta = np.linalg.pinv(xmt_wm @ xm) @ xmt_wm @ ym
            y = (beta @ xp)[0]
        else:
            xx = self.n_xx[min_range]
            yy = self.n_yy[min_range]
            sum_weight = np.sum(weights)
            sum_weight_x = np.dot(xx, weights)
            sum_weight_y = np.dot(yy, weights)
            sum_weight_x2 = np.dot(np.multiply(xx, xx), weights)
            sum_weight_xy = np.dot(np.multiply(xx, yy), weights)

            mean_x = sum_weight_x / sum_weight
            mean_y = sum_weight_y / sum_weight

            b = (sum_weight_xy - mean_x * mean_y * sum_weight) / \
                (sum_weight_x2 - mean_x * mean_x * sum_weight)
            a = mean_y - b * mean_x
            y = a + b * n_x
        return self.denormalize_y(y)


def main():
    xx = np.array([0.5578196, 2.0217271, 2.5773252, 3.4140288, 4.3014084,
                   4.7448394, 5.1073781, 6.5411662, 6.7216176, 7.2600583,
                   8.1335874, 9.1224379, 11.9296663, 12.3797674, 13.2728619,
                   14.2767453, 15.3731026, 15.6476637, 18.5605355, 18.5866354,
                   18.7572812])
    yy = np.array([18.63654, 103.49646, 150.35391, 190.51031, 208.70115,
                   213.71135, 228.49353, 233.55387, 234.55054, 223.89225,
                   227.68339, 223.91982, 168.01999, 164.95750, 152.61107,
                   160.78742, 168.55567, 152.42658, 221.70702, 222.69040,
                   243.18828])

    loess = Loess(xx, yy)

    with open("data.csv", "w") as f:
        for x, y in zip(xx, yy):
            f.write(f"{x}, {y}\n")

    with open("loess.csv", "w") as f:
        for x in np.linspace(xx.min(), xx.max(), num=100):
            e7 = loess.estimate(x, window=7, use_matrix=False, degree=1)
            e14 = loess.estimate(x, window=14, use_matrix=False, degree=1)
            f.write(f"{x}, {e7}, {e14}\n")


def plot():
    import csv
    import matplotlib.pyplot as plt

    with open("data.csv", newline="") as f:
        reader = csv.reader(f)
        data = np.array([row for row in reader]).astype("f8")

    with open("loess.csv", newline="") as f:
        reader = csv.reader(f)
        smoothed = np.array([row for row in reader]).astype("f8")

    xx = data[:, 0]
    yy = data[:, 1]

    xs = smoothed[:, 0]
    e7 = smoothed[:, 1]
    e14 = smoothed[:, 2]

    plt.scatter(xx, yy, s=4, label="data")
    plt.plot(xs, e7, "r", label="Window: 7")
    plt.plot(xs, e14, "m", label="Window: 14")

    plt.legend()
    plt.show()


if __name__ == "__main__":
    start = time.time()

    plot()

    end = time. time()
    print(end - start)



def demand_loop(building,profile,T_mains,node,l):
    Cp_water = 4186  #J/kg*K
    equip = building.plant_demand_equip 
    on_loop = [i for i,x in enumerate(equip['loop']) if x == l] 
    for j in on_loop:
        in_n = equip['inlet'][j] 
        out_n = equip['outlet'][j] 
        if equip['type'][j]=='pipe:adiabatic':
            node['demand_temperature'][out_n] = node['demand_temperature'][in_n] 
        elif equip['type'][j]=='mixer':
            try:
                in_n[0]
            except TypeError:
                in_n = [in_n]
            if node['demand_flow'][out_n]>0:
                node['demand_temperature'][out_n] = sum([node['demand_temperature'][k]*node['demand_flow'][k]/node['demand_flow'][out_n]  for k in in_n])
            else: #avoid divide by zero
                node['demand_temperature'][out_n] = sum([node['demand_temperature'][k] for k in in_n])/len(in_n) 
        elif equip['type'][j]=='splitter':
            try:
                out_n[0]
            except TypeError:
                out_n = [out_n]
            for k in out_n:
                node['demand_temperature'][k] = node['demand_temperature'][in_n] 
        elif (equip['type'][j]=='coil:cooling:water' or
            equip['type'][j]=='coil:cooling:water:detailedgeometry' or
            equip['type'][j]=='coil:heating:water' or
            equip['type'][j]=='chiller:electric:reformulatedeir' or
            equip['type'][j]=='chiller:electric:eir'):
            if node['demand_flow'][out_n]>0:
                node['demand_temperature'][out_n] = node['demand_temperature'][in_n] + node['load'][out_n]/(Cp_water*node['demand_flow'][out_n]) 
            else:#avoid divide by zero
                node['demand_temperature'][out_n] = node['demand_temperature'][in_n] 
        elif equip['type'][j]=='wateruse:connections':
            try:
                node['demand_temperature'][out_n] = profile[building.water_use[equip['name'][j] ]['cold_supply_temperature_schedule']]
            except KeyError:
                node['demand_temperature'][out_n] = T_mains            
        else:
            print('need additional plant demand loop component of type__' + equip['type'][j])
    return node


'''

'''
from eagers.building.basic.psychometric import psychometric
from eagers.building.basic.two_way_interp import two_way_interp

from scipy.special import j0 as besseli0
from scipy.special import j1 as besseli1
from scipy.special import kn as besselk

import numpy as np

def detailed_water_coil(coil,air_in,sensible_load,Tw_in,T_supply):
    air_out = {}
    for k in air_in:
        air_out[k] = air_in[k]+0
    if air_in['m_dot']>0 and sensible_load>0:
        den_water = 995 #kg/m^3
        pi = 3.14159
        Cp_air = 1006 + 1860*air_in['w'] #J/kg*K
        m_w_rated = coil['rated_water_flow']*den_water #convert m^3/s to kg/s
        h1 = psychometric({'Twb':Tw_in},'h')
        h2 = psychometric({'Twb':Tw_in+1},'h')
        bb = h2-h1
        aa = h1 - bb*Tw_in
        Pr = 0.733
        viscosity = 1.846e-5
        tube_thickness = (coil['tube_od']-coil['tube_id'])/2
        pipe_area = coil['tubes_per_row']*pi*(coil['tube_id']**2)/4
        surface_area = coil['tube_area_outside'] + coil['fin_area'] #total_coil_outside_area
        fin_height = (coil['fin_diameter']-coil['tube_od'])/2
        fin_effective_diam = (4*coil['fin_diameter']*coil['coil_depth']/(pi*coil['tube_rows']*coil['tubes_per_row']))**.5
        FA2SA = coil['fin_area']/surface_area #% Ratio of secondary (fin) to total (secondary plus primary) surface areas
        ma_scaled = (1+air_in['w'])*air_in['m_dot']/coil['min_airflow_area']
        coil_id_effective = 4*coil['min_airflow_area']*coil['coil_depth']/surface_area
        Re_a = coil_id_effective*ma_scaled/viscosity #Re_a should be between 400 and 1500
        tube_to_fin = coil['tube_od']/fin_effective_diam
        fin_diam_var = 0.5*(fin_effective_diam - coil['tube_od'])
        C1 = 0.159*(coil['fin_thickness']/coil_id_effective)**(-0.065)*(coil['fin_thickness']/fin_diam_var)**0.141
        C2 = -0.323*(coil['fin_spacing']/fin_diam_var)**0.049*(fin_effective_diam/coil['tube_spacing'])**0.549*(coil['fin_thickness']/coil['fin_spacing'])**(-0.028)

        f_o = C1*Re_a**C2*ma_scaled*Cp_air*Pr**(-2/3)
        f_o_w = f_o*(1.425 - 5.1e-4*Re_a + 2.63e-7*Re_a**2)
        
        n_o = surface_efficiency(f_o,tube_to_fin,coil['tube_od'],FA2SA,fin_effective_diam,coil['fin_conductivity'],coil['fin_thickness'])
        n_o_w = surface_efficiency(f_o_w,tube_to_fin,coil['tube_od'],FA2SA,fin_effective_diam,coil['fin_conductivity'],coil['fin_thickness'])
        
        R_o = 1/(f_o*n_o*surface_area)
        R_o_w = Cp_air/bb/(f_o_w*n_o_w*surface_area)
        R_mf = tube_thickness/(coil['tube_conductivity']*coil['tube_area_inside']) # + 5e-2/coil.tube_area_inside(k) #fouling factor from page 800 is 5e-2 m^2*K/W
        ## line search mass flow of water to get to desired outlet temperature
        n = 10
        Cp_water = 4186 #J/kg*K
        Tout = [0 for i in range(n)]
        Tw_out = max(coil['water_outlet_temperature'],Tw_in + 1)
        m_w_nom = max(1e-8*m_w_rated,sensible_load/(Cp_water*(Tw_out - Tw_in)))
        m_w_test = np.logspace(np.log10(.12*m_w_nom),np.log10(m_w_rated),n)
        mw_test = [float(x) for x in m_w_test]
        for i in range(n):
            Tout[i],_ = detailed_coil_heat_transfer(m_w_test[i],air_in,Tw_in,pipe_area,coil['tube_area_inside'],coil['tube_id'],aa,bb,R_mf,R_o,R_o_w)
        if Tout[-1]>T_supply:
            m_w = m_w_rated
        elif Tout[0]<T_supply:
            m_w = m_w_test[0]
        else:
            m_w = two_way_interp(T_supply,Tout,m_w_test)
        air_out['Tdb'],air_out['w'] = detailed_coil_heat_transfer(m_w,air_in,Tw_in,pipe_area,coil['tube_area_inside'],coil['tube_id'],aa,bb,R_mf,R_o,R_o_w)
        air_out['h'] = psychometric(air_out,'h')
        T_water_out = Tw_in + (-air_out['m_dot']*(air_out['h'] - air_in['h']))/(Cp_water*m_w)
    actual_load = -air_out['m_dot']*(air_out['h'] - air_in['h']) #heat added in W
    return air_out,m_w,actual_load


def surface_efficiency(f_o,rho,tube_od,FA2SA,fin_diameter,fin_conductivity,fin_thickness):
    fai = (fin_diameter - tube_od)/2*(2*f_o/(fin_conductivity*fin_thickness))**.5
    u_e = fai/(1-rho)
    u_b = u_e*rho
    n_fin = -2*rho/(fai*(1+rho))*(besseli1(u_b)*besselk(1,u_e) - besselk(1,u_b)*besseli1(u_e))/(besseli0(u_b)*besselk(1,u_e) - besselk(0,u_b)*besseli1(u_e))
    n_o = 1 - (1 - n_fin)*FA2SA
    return n_o

def detailed_coil_heat_transfer(m_w,air_in,Tw_in,pipe_area,area_inside,tube_id,aa,bb,R_mf,R_o,R_o_w):
    den_water = 995 #kg/m^3
    Cp_water = 4186 #J/kg*K
    Cp_air = 1006 + 1860*air_in['w'] #J/kg*K

    V_w = m_w/den_water/pipe_area #velocity in m/s
    water_tube_thermresist =(tube_id)**0.2/(area_inside*1.429*(V_w**0.8)) # 
    f_i = (1+0.0146*(Tw_in+273))/water_tube_thermresist
    R_i = 1/(f_i*area_inside)
    UA_dry = (1/(R_i + R_mf + R_o))
    ## compute exit temperature with all dry
    X = 1/(air_in['m_dot']*Cp_air)
    Y = -1/(m_w*Cp_water)
    Z_d = float(np.exp(UA_dry*(X + Y)))
    K1 = (Z_d-1)/(Z_d +Y/X)
    # K1 = X*(Z-1)/(X*Z+Y)
    # K2 = (X+Y)*Z/(X*Z+Y)
    Tw_2 = Tw_in
    Ta_2 = air_in['Tdb'] - K1*(air_in['Tdb'] - Tw_2)
    ## Check if exit dewpoint temperature is less than surface temperature
    Tdp = psychometric(air_in,'Tdp')
    T_s_2 = Tw_2 + ((1/UA_dry)-R_o)/(1/UA_dry)*(Ta_2 - Tw_2)
    if Tdp<T_s_2:
        ## completely dry coil
        Ta_out = Ta_2
        w_out = air_in['w']
    else:
        ## not completely dry, line search to find portion wet
        X2 = 1/air_in['m_dot']
        Y2 = -bb/(m_w*Cp_water)
        x_test = np.logspace(-2,0,25)
        x_test = [float(x) for x in x_test]
        T_s_2 = [0 for i in range(len(x_test)+1)]
        T_s_2[0],_,_ = wet_coil_calc(1e-4,aa,bb,R_i,R_mf,R_o,R_o_w,UA_dry,X,Y,X2,Y2,air_in,Tw_in)
        for i in range(len(x_test)):
            T_s_2[i+1],_,_ = wet_coil_calc(x_test[i],aa,bb,R_i,R_mf,R_o,R_o_w,UA_dry,X,Y,X2,Y2,air_in,Tw_in)
        x = two_way_interp(Tdp,T_s_2,[1e-4]+x_test)
        _,Ta_2,Tw_2 = wet_coil_calc(x,aa,bb,R_i,R_mf,R_o,R_o_w,UA_dry,X,Y,X2,Y2,air_in,Tw_in)

        Q_w = m_w*Cp_water*(Tw_2 - Tw_in)
        Ha_2 = psychometric({'Tdb':Ta_2,'w':air_in['w']},'h') #condition at end of dry section (equal to dew point)
        Ha3 = Ha_2 - Q_w/air_in['m_dot']
        Ta_out = psychometric({'h':Ha3},'Tdb')
        w_out = psychometric({'h':Ha3,'Tdb':Ta_out},'w')
    return Ta_out,w_out

def wet_coil_calc(x,aa,bb,R_i,R_mf,R_o,R_o_w,UA_dry,X,Y,X2,Y2,air_in,Tw_in):
    Cp_air = 1006 + 1860*air_in['w'] #J/kg*K
    if x == 0: #fully wet 
        UA_wet = (1/bb/(R_i + R_mf + R_o_w))
        Z_w = float(np.exp(UA_wet*(X2 + Y2)))
        #  K3 = (X2+Y2)/(X2*Z_w + Y2)
        K5 = Z_w*(Y2 + X2)/(X2*Z_w + Y2)
        K6 = Y2*(1 - Z_w)/(bb*(X2*Z_w + Y2))
        Tw_2 = K5*Tw_in + K6*(air_in['h'] - aa)
        Ta_2 = air_in['Tdb']
        R_x = (R_mf + R_i)/R_o_w
        T_s_2 = 1/(R_x+1)*(Tw_2 + R_x/bb*(air_in['h'] - aa))
    else:#partially wet
        UA_wet = (1-x)*(1/bb/(R_i + R_mf + R_o_w))
        UA_dry_x = x*UA_dry
        Z_d = float(np.exp(UA_dry_x*(X + Y)))
        Z_w = float(np.exp(UA_wet*(X2 + Y2)))
        K1 = (Z_d-1)/(Z_d +Y/X)
        K5 = Z_w*(Y2 + X2)/(X2*Z_w + Y2)
        K6 = Y2*(1 - Z_w)/(bb*(X2*Z_w + Y2))
        Tw_2 = 1/(1 - K1*K6*Cp_air)*(K5*Tw_in + K6*(air_in['h'] - Cp_air*air_in['Tdb']*K1 -aa))
        Ta_2 = air_in['Tdb'] - K1*(air_in['Tdb'] - Tw_2)
        T_s_2 = Tw_2 + ((1/UA_dry_x)-R_o)/(1/UA_dry_x)*(Ta_2 - Tw_2)
    return T_s_2,Ta_2,Tw_2


"""File handling utilities.

Functions:
ensure_dirpath
ensure_nonduplicate_with_version
ensure_suffix
find_file_in_tree
find_file_in_userdir
"""

import os
import pathlib

from eagers.config.user import USER_DIR_EXCLUDED
from eagers.config.text import NAME_SEP, VERSION_INDICATOR


def ensure_dirpath(filepath):
    """Returns the given filepath after creating the necessary
    directories for it to exist, if needed.

    Positional arguments:
    filepath - (pathlib.Path) Path to file or directory.
    """
    try:
        filepath.mkdir(parents=True, exist_ok=True)
    except FileExistsError:
        # Last path component is an existing non-directory file (see
        # pathlib.Path.mkdir() documentation).  Ensure its parent
        # exists.
        filepath.parent.mkdir(parents=True, exist_ok=True)
    except AttributeError:
        raise TypeError(f"{filepath!r} is not a pathlib.Path instance.")
    return filepath


def ensure_nonduplicate_with_version(filepath, version=None):
    """Return the given filepath with a new name, if necessary,
    including a version suffix to ensure it is not a duplicate of an
    already-existing file.

    Positional arguments:
    filepath - (pathlib.Path) Path to file.

    Keyword arguments:
    version - (int) (Default: None) Version number to use. Allows manual
        specification of a version number.
    """
    def new_filepath(filepath, vn):
        """Return a new filepath with the specified version number."""
        stem, suffix = filepath.stem, filepath.suffix
        return filepath.with_name(
            f"{stem}{NAME_SEP}{VERSION_INDICATOR}{vn}" + suffix
        )
    if version is not None:
        # Version manually specified.
        filepath = new_filepath(filepath, version)
        if filepath.exists():
            raise ValueError(
                f"Version {version!r} does not create a unique filepath."
            )
    if not filepath.exists():
        # Job's done. No need to worry about versions.
        return filepath
    # Add a version suffix and increment the version number until there
    # is no more duplication.
    vnum = 2
    while filepath.exists():
        filepath = new_filepath(filepath, vnum)
        vnum += 1
    return filepath


def ensure_suffix(filepath, suffix, remove_bad_suffix=False):
    """Return the given file path, ensuring that it has the given
    suffix. The returned file path has the same data type as the given
    one.

    Positional arguments:
    filepath - (pathlib.Path or str) The file path.
    suffix - (str) The suffix.

    Keyword arguments:
    remove_bad_suffix - (bool) (Default: False) Whether to replace the
        current with the correct one, or to simply append the correct
        one. WARNING: Setting this to True can result in unexpected
        behavior if the filename contains dots.
    """
    # Make sure suffix has a dot at the beginning.
    if suffix[0] != '.':
        suffix = '.' + suffix
    # Make sure pathlib.Path is used.
    str_input = isinstance(filepath, str)
    filepath = pathlib.Path(filepath)
    # Check suffix.
    if filepath.suffix != suffix:
        if remove_bad_suffix:
            filepath = filepath.with_name(filepath.stem + suffix)
        else:
            filepath = filepath.with_name(filepath.name + suffix)
    return str(filepath) if str_input else filepath


def find_file_in_tree(dir_, filename, exclude_dirs=None):
    """Find the given file name in the file tree rooted at the given
    directory. Returns the path to the file. If the file does not exist,
    a ValueError is raised. If more than one instance of the file
    exists, a RuntimeError is raised.

    Positional arguments:
    dir_ - (pathlib.Path) Path to tree root directory.
    filename - (str) Name of file.

    Keyword arguments:
    exclude_dirs - (list of str) Directories to exclude from search.
    """
    # Handle input.
    if not exclude_dirs:
        exclude_dirs = []
    # Search directory tree.
    found = False
    found_dir = None
    for root, dirs, files in os.walk(ensure_dirpath(dir_)):
        if root.rsplit(os.sep, maxsplit=1)[-1] in exclude_dirs:
            # dirs list must be modified in-place; not reassigned.
            for i in range(len(dirs)):
                del dirs[0]
        elif filename in files:
            if found:
                dirname = lambda d: d.rsplit(os.path.sep, maxsplit=1)[1]
                raise RuntimeError(
                    f"{filename!r} found in at least two directories: "
                    f"{dirname(found_dir)!r}, {dirname(root)!r}")
            found = True
            found_dir = root
    if not found:
        raise ValueError(f"{filename!r} not in {dir_.name!r}")
    return pathlib.Path(found_dir) / filename


def find_file_in_userdir(subdir, filename, suffix):
    """Find the file with the given name in the given sub-directory of
    the user directory.

    Positional arguments:
    subdir - (pathlib.Path) Sub-directory of user directory.
    filename - (str) File name.
    suffix - (str) File suffix.
    """
    return find_file_in_tree(
        subdir,
        ensure_suffix(filename, suffix),
        exclude_dirs=USER_DIR_EXCLUDED,
    )



"""Test cases with fewest number of generators, then sort others by
estimated cost.

Functions:
test_min_cases
build_cases
list_cases
validate_combinations
check_feas
group_gen
find_last_group
split_to_groups
find_avg_kWh_cost
"""

from eagers.solver.test_cases import test_cases
from eagers.building.basic.two_way_interp import two_way_interp


def test_min_cases(qp, gen, options, demand, c_red, step_profile, stor_power, dt):
    """Test cases with fewest number of generators, then sort other by
    estimated cost.

    INPUTS:
    qp:             qp instance of quadratic programming matrices.
    gen:            List of generator objects.
    options:        Optimization options.
    demand:         Forecasted demand minus the forecasted renewable
                        generation plus the minimum building energy load.
    step_profile:  The first estimate of generator dispatch.
    dt:             Delta time vector.

    OUTPUTS:
    verified:       Which combination of unit commitment is verified
                        feasible.
    cost:           Vector of costs for each combination.
    combinations:      Unit commitment matrix.
    disp_comb:      
    """

    # TODO: Add parallel computing logic.
    parallel = False
    
    combinations,n_gen = build_cases(qp,gen,options,demand,step_profile,dt,c_red,stor_power,parallel)
    binary_comb = [[j>0 for j in combinations[i]] for i in range(len(combinations))]
    # if parallel and n_gen!=None:
    #     verified,cost,disp_comb,heating,cooling,build_temp,fluid_temp = test_cases_parallel(qp,gen,combinations,n_gen,c_red,stor_power,None)
    # elif n_gen!=None:
    verified,cost,disp_comb,heating,cooling,build_temp,fluid_temp = test_cases(qp,gen,combinations,n_gen,c_red,stor_power,None) 
    return verified,cost,binary_comb,disp_comb,heating,cooling,build_temp,fluid_temp


def build_cases(qp, gen, options, demand, step_profile, dt,c_red,stor_power,parallel):
    n_g = len(gen)
    always_on = []
    p_gen = {'e': [], 'h':[], 'c':[], 'cw':[], 'hy':[]}
    for i, g in enumerate(gen):
        if g['type'] in ['ElectricGenerator','CombinedHeatPower']:
            p_gen['e'].append(i)
        elif g['type'] == 'Heater':
            p_gen['h'].append(i)
        elif g['type'] in ['Chiller','AbsorptionChiller']:
            p_gen['c'].append(i)
        elif g['type'] in ['CoolingTower']:
            p_gen['cw'].append(i)
        elif g['type'] in ['HydrogenGenerator','Electrolyzer',]:
            p_gen['hy'].append(i)
        elif qp['organize']['enabled'][i]:
            always_on.append(i)
    l_type = [len(p_gen[k]) for k in p_gen]
    if any([x>4 for x in l_type]) and sum(l_type)>8:
        combinations,n_gen = group_gen(p_gen,always_on,qp,gen,options,demand,step_profile,dt,c_red,stor_power,parallel)
    else:
        combinations,n_gen = list_cases(p_gen,always_on,n_g) #a matrix of all the possible generator on/off combinations
        validate_combinations(combinations,n_gen,qp,gen,options,demand,step_profile,dt)
    return combinations, n_gen

def list_cases(p_gen,always_on,n_g):
    def mult_list(val):
        ans = 1
        for k in val:
            ans *= k
        return ans
    def binary_dispatchable(comb,k,z):
        r = 0
        while r+z <= len(comb):
            for rc in range(r,r+z):
                comb[rc][k] = True
            r += 2*z
        return comb
    def group_dispatchable(comb,k,z):
        r = 0
        n = len(k)
        while r+n*z <= len(comb):
            r+=z
            r_last = r+n*z
            for j in k:
                for rc in range(r,r_last):
                    for i in j:
                        comb[rc][i] = True
                r += z
        return comb
    c_mult =[]
    for type_gen in p_gen:
        g_list = p_gen[type_gen]
        if len(g_list)>0:
            if isinstance(g_list[0], (float,int)) or len(g_list[0])==1: #unsorted components test all combinations
                c_mult.extend([2 for i in range(len(g_list))])
            else: #sorted sequence
                c_mult.append(len(g_list)+1)
    c_len = mult_list(c_mult)
    base = [False for j in range(n_g)]
    for k in always_on:
        base[k] = True
    combinations = [[j for j in base] for i in range(c_len)]

    e = 1
    for type_gen in p_gen:
        g_list = p_gen[type_gen]
        if len(g_list)>0:
            if isinstance(g_list[0], list) and len(g_list[0])==1:
                g_list = [g[0] for g in g_list]
            if isinstance(g_list[0], (float,int)):
                for k in range(len(g_list)):
                    z = mult_list(c_mult[k+e:])
                    combinations = binary_dispatchable(combinations,g_list[k],z)
                e += len(g_list)
            else:
                z = mult_list(c_mult[e:])
                combinations = group_dispatchable(combinations,g_list,z)
                e += 1
     
    active_gen = [sum(combinations[i])-len(always_on) for i in range(c_len)]
    reorder = sorted(range(len(active_gen)), key=lambda k: active_gen[k])
    n_gen = [active_gen[i] for i in reorder]
    sorted_comb = [combinations[i] for i in reorder]
    return sorted_comb, n_gen

def validate_combinations(combinations, n_gen, qp, gen, options, demand, step_profile, dt):
    #TODO validate with node based loads and transmission limits
    ac_dc = None
    for i in range(len(gen)):
        if gen[i]['type'] == 'ACDCConverter':
            ac_dc = i
            break
    for net in list(qp['organize']['balance'].keys()):
        gen_dem = 0
        if net in demand:
            gen_dem = sum(demand[net][k] for k in demand[net])
        r_eq = qp['organize']['balance'][net] #rows of Aeq associated with electric demand

        min_prod, max_prod = check_feas(gen, qp, r_eq, step_profile, dt, net)
        if net in ['electrical','direct_current'] and not ac_dc is None:
             min_prod[i], max_prod[i] = ac_dc_prod(net, qp, r_eq[gen[ac_dc]['subnet_node'][net]], qp['indices'][ac_dc][0], ac_dc, gen, step_profile, dt)
        # Keep the rows where the ub is capable of meeting demand and
        # the lower bound is low enough to meet demand.
        r = 0
        while r<len(combinations):
            del_row = False
            if (net == 'district_heat' and options['excess_heat']) or (net == 'district_cool' and options['excess_cool']):
                pass #no lower bound constraint
            elif sum([min_prod[k] for k in range(len(gen)) if combinations[r][k]]) > gen_dem:
                del_row = True
            if del_row or sum([max_prod[k] for k in range(len(gen)) if combinations[r][k]]) < gen_dem:
                del combinations[r]
                del n_gen[r]
            else:
                r+=1
        if len(combinations)==0:
            print('Zero feasible test combinations')
    # return combinations, n_gen

def check_feas(gen, qp, r_eq, step_profile, dt, net):
    """Returns the minimum and maximum production on the network
    specified by net, for each combination of generators.
    """
    n_g = len(gen)
    min_prod = [0] * n_g
    max_prod = [0] * n_g

    for i, g in enumerate(gen):
        states = qp['indices'][i][0]
        if not states is None and net in gen[i]['subnet_node']: #any([any([qp['a_eq'][r][s]!=0 for s in states]) for r in r_eq]): #2nd condition checks for qp.a_eq at index [r_eq,states] != 0
            row_i = r_eq[gen[i]['subnet_node'][net]]
            if gen[i]['type'] == 'Utility':
                if len(states) == 2:
                    min_prod[i] = qp['a_eq'][row_i][states[1]]* qp['ub'][states[1]]
                    max_prod[i] = qp['a_eq'][row_i][states[0]]* qp['ub'][states[0]]
                else:
                    min_prod[i] = qp['a_eq'][row_i][states[0]] * qp['lb'][states[0]]
                    max_prod[i] = qp['a_eq'][row_i][states[0]] * qp['ub'][states[0]]
            elif 'Storage' in g['type']:
                s = states[0]
                min_prod[i] = -g['ramp']['b'][0]
                max_prod[i] = g['ramp']['b'][1]
                if step_profile is not None:
                    max_prod[i] = min([step_profile[i] * qp['a_eq'][row_i][s] / dt, max_prod[i]]) 
                    charging_space = (g['stor']['usable_size'] -step_profile[i])* qp['a_eq'][row_i][s]
                    min_prod[i] = max([-charging_space/dt, min_prod[i]])
            elif gen[i]['type'] == 'ACDCConverter':
                continue #done later
            elif gen[i]['type'] in ['ElectricGenerator','CombinedHeatPower','Heater','Chiller','AbsorptionChiller','CoolingTower','HydrogenGenerator','Electrolyzer']:
                if qp['a_eq'][row_i][states[0]] > 0:
                    min_prod[i] = sum([qp['a_eq'][row_i][s] * qp['lb'][s] for s in states])
                    max_prod[i] = sum([qp['a_eq'][row_i][s] * qp['ub'][s] for s in states])
                else:
                    # If it consumes this demand (i.e. a chiller when examining electrical demand) then the lower bound is -1*ub.
                    min_prod[i] = sum([qp['a_eq'][row_i][s] * qp['ub'][s] for s in states])
                    max_prod[i] = sum([qp['a_eq'][row_i][s] * qp['lb'][s] for s in states])
            if 'const_demand' in gen[i] and net in gen[i]['const_demand']:
                max_prod[i] -= gen[i]['const_demand'][net]
                min_prod[i] -= gen[i]['const_demand'][net]
    return min_prod, max_prod

def ac_dc_prod(net, qp, row, states, ac_dc, gen, step_profile, dt):
    if net == 'electrical':
        row_dc = qp['organize']['balance']['direct_current']
        _, max_prod_dc = check_feas(gen, qp, row_dc, step_profile, dt,'direct_current')
        tot_dc = sum([max_prod_dc[k] for k in range(len(gen))])
        # Capacity to convert AC to DC power.
        min_prod = -gen[ac_dc]['a']['ub']
        # Pull out the DC/AC conversion efficiency.
        max_prod = qp['a_eq'][row][states[1]] * tot_dc
    elif net == 'direct_current':
        # Need only the row corresponding to the node that this generator is on.
        row_ac = qp['organize']['balance']['electrical']
        _, max_prod_ac = check_feas(gen, qp, row_ac, step_profile, dt,'electrical')
        tot_ac = sum([max_prod_ac[k] for k in range(len(gen))])
        # Capacity to convert DC to AC power.
        min_prod = -gen[ac_dc]['b']['ub']
        # Pull out the AC/DC conversion efficiency.
        max_prod = qp['a_eq'][row][states[0]] * tot_ac
    return min_prod, max_prod

def group_gen(p_gen,always_on,qp,gen,options,demand,step_profile,dt,c_red,stor_power,parallel):
    ## with more than 1000 combinations, group generators to reduce the search parameters

    #group by similar average cost per kW by taking samples across range
    #when testing, the lower limit of the most expensive group gets set to
    #zero, except for the smallest lower limit in the group
    #non-electric generators (chillers) don't get grouped
    p_groups = {}
    for type_gen in p_gen:
        g_list = p_gen[type_gen]
        if len(g_list)>4:
            avg_cost_kWh = []
            for k in g_list:
                states = qp['indices'][k][0]
                h_gen = [qp['h'][s] for s in states]
                avg_cost_kWh.append(find_avg_kWh_cost(h_gen,[qp['f'][s] for s in states],[qp['lb'][s] for s in states],[qp['ub'][s] for s in states],qp['const_cost'][k],gen[k],c_red['c'][k]))
            
            gen_index = sorted(range(len(avg_cost_kWh)), key=lambda k: avg_cost_kWh[k])
            sorted_gen = [g_list[i] for i in gen_index]
            p_groups[type_gen] = split_to_groups(sorted_gen)
        else:
            p_groups[type_gen] = split_to_groups(g_list)

    #create cases with a) group 1, b) group 1 & 2, c) group 1 & 2 & 3 ....
    combinations,n_gen = list_cases(p_groups,always_on,len(gen)) #a matrix of all the possible generator on/off combinations
    validate_combinations(combinations,n_gen,qp,gen,options,demand,step_profile,dt)
    ## TODO add parallel computing for test_cases
    # if parallel and len(n_gen)>0
    #     _,cost,_,_,_,_,_  = test_cases_parallel(qp,gen,combinations,n_gen,c_red,p_groups)
    # elif  len(n_gen)>0
    _,cost,_,_,_,_,_ = test_cases(qp,gen,combinations,n_gen,c_red,stor_power,p_groups)

    #Find group that is on the bubble
    best_combo = combinations[min(range(len(cost)), key=cost.__getitem__)]
    for k in list(p_gen.keys()):
        if len(p_gen[k])>4:
            p_gen[k], always_on  = find_last_group(best_combo,p_groups[k],always_on)    
    
    #Break apart this group and make combinations based only on it
    combinations,n_gen= list_cases(p_gen,always_on,len(gen)) #a matrix of all the possible generator on/off combinations
    validate_combinations(combinations,n_gen,qp,gen,options,demand,step_profile,dt)
    return  combinations,n_gen


def find_last_group(best_combo,groups,always_on):
    #find last group that has an active gen
    k = 0
    while k<len(groups)-1 and any([best_combo[j] is True for j in groups[k+1]]):
        if k>0:
            always_on.extend(groups[k-1]) #cheaper generators are held on
        k +=1
    obj_list = []
    if k>0:
        obj_list.extend(groups[k-1])
    obj_list.extend(groups[k])
    return obj_list, always_on


def split_to_groups(obj_list):
    #split into groups
    n_inc = len(obj_list)
    if n_inc<=4:
        groups = [[x] for x in obj_list]
        # groups = obj_list
    elif n_inc<=6:
        # groups = [obj_list[:2],obj_list[2:4],obj_list[4:]]
        groups = [obj_list[:3],obj_list[3:]]
    else:
        n = 4
        n_group = round(n_inc/n)
        groups =[]
        for i in range(n_group):
            if (i+1)*n>n_inc:
                groups.append(obj_list[i*n:])
            else:
                groups.append(obj_list[i*n:(i+1)*n])
    return groups


def find_avg_kWh_cost(H,f,lb,ub,const,gen,min_load_cost):
    n = 10
    cost_kWh = []
    max_p = sum(ub)
    capacity = [lb[0]+i*(max_p-lb[0])/(n-1) for i in range(n)]
    if all([H[j]== 0 for j in range(len(H))]) and all([f[j]== 0 for j in range(len(H))]):
        load = True
    else:
        load = False
    if capacity[0] == 0:
        del capacity[0]
        n -=1
    for i in range(n):
        j = 0
        p = 0
        cost = const
        cap = capacity[i]
        while p<cap and j<len(ub):
            x = min(cap-p,ub[j])
            if load:
                eff = two_way_interp(p+x,gen['cap'],gen['eff'])
                cost += x*min_load_cost/eff
            else:
                cost += x**2*H[j] + x*f[j]
            p = p+x
            j = j+1
        cost_kWh.append(cost/capacity[i])
    avg_cost_kWh = sum(cost_kWh)/n
    return avg_cost_kWh
